from pyxbos.process import XBOSProcess, b64decode, b64encode, schedule, run_loop
from pyxbos.xbos_pb2 import XBOS
from pyxbos.nullabletypes_pb2 import Double
from pyxbos.energise_pb2 import EnergiseMessage, LPBCStatus, LPBCCommand, SPBC, EnergiseError, EnergisePhasorTarget, ChannelStatus
from pyxbos.c37_pb2 import Phasor, PhasorChannel
from datetime import datetime
from functools import partial
from collections import deque
import asyncio
import traceback

class ConfigMissingError(Exception): pass

class SPBCProcess(XBOSProcess):
    """
    Wrapper process for supervisory phasor-based control in Python.

    An SPBC subscribes to a set of LPBCs and receives status mesages from them.
    These status messages consist of an error quantity and saturation state. 
    In the current implementation, the SPBC subscribes to all LPBCs that it has
    permission to see.

    At regular intervals, the SPBC publishes a V + delta target for *each* node,
    represented by an LPBC.
    """

    def __init__(self, cfg):
        super().__init__(cfg)
        if 'namespace' not in cfg:
            raise ConfigMissingError('namespace')
        if 'name' not in cfg:
            raise ConfigMissingError('name')
        if 'reference_channels' not in cfg:
            raise ConfigMissingError('reference_channels')
        self.namespace = b64decode(cfg['namespace'])
        self._log.info(f"initialized SPBC: {cfg}")
        self.name = cfg['name']

        # reference channels are URIs for the uPMU channels the SPBC
        # subscribes to. The SPBC framework maintains self.reference_phasors
        # to contain the most recent phasor measurements for each channel
        self._reference_channels = cfg['reference_channels']
        self.reference_phasors = {k: None for k in self._reference_channels}
        for channel in self._reference_channels:
            upmu_uri = f"upmu/{channel}"
            self._log.info(f"Subscribing to {channel} as reference phasor")
            schedule(self.subscribe_extract(self.namespace, upmu_uri, ".C37DataFrame", self._upmucb, "spbc_reference"))

        self.lpbcs = {}
        schedule(self.subscribe_extract(self.namespace, "lpbc/*", ".EnergiseMessage.LPBCStatus", self._lpbccb, "lpbc_status"))

    def _upmucb(self, resp):
        """
        Called on every received C37 frame from the reference upmu channels.
        Stores the most recent phasors received:

        self.reference_phasors = {
            'flexlab1/L1': [
                {
                    "time": "1559231114799996800",
                    "angle": 193.30149788923268,
                    "magnitude": 0.038565948605537415
                },
                {
                    "time": "1559231114899996400",
                    "angle": 195.50249902851263,
                    "magnitude": 0.042079225182533264
                }
                ... etc
            ],
            'flexlab/L2': [
                {
                    "time": "1559231114799996800",
                    "angle": 220.30149788923268,
                    "magnitude": 10.038565948605537415
                },
                {
                    "time": "1559231114899996400",
                    "angle": 220.50249902851263,
                    "magnitude": 10.042079225182533264
                }
            ]
        }
        """
        upmu = resp.uri[5:]
        self.reference_phasors[upmu] = resp.values[-1]['phasorChannels'][0]['data']

    def _lpbccb(self, resp):
        """
        Caches the last message heard from each LPBC

        Each LPBC status looks like:
        {
            # local time of LPBC
            'time': 1559231114799996800,
            # phasor errors of LPBC
            'phasor_errors': {
                'angle': 1.12132,
                'magnitude': 31.12093090,
                # ... and/or ...
                'P': 1.12132,
                'Q': 31.12093090,
            },
            # true if P is saturated
            'pSaturated': True,
            # true if Q is saturated
            'qSaturated': True,
            # if pSaturated is True, expect the p max value
            'pMax': 1.4,
            # if qSaturated is True, expect the q max value
            'qMax': 11.4,
            # true if LPBc is doing control
            'do_control': True,
        }
        """
        statuses = resp.values[-1]
        timestamp = statuses['time']
        for status in statuses['statuses']:
            if status['nodeID'] not in self.lpbcs:
                self.lpbcs[status['nodeID']] = {}
            if 'pSaturated' not in status:
                status['pSaturated'] = False
            if 'qSaturated' not in status:
                status['qSaturated'] = False
            if 'pMax' in status:
                status['pMax'] = status['pMax']['value']
            if 'qMax' in status:
                status['qMax'] = status['qMax']['value']
            self.lpbcs[status['nodeID']][status['channelName']] = status
        #self.lpbcs[resp.uri] = resp

    async def broadcast_target(self, nodeid, channels, vmags, vangs, kvbases=None, kvabases=None):
        """
        Publishes SPBC V and delta for a particular node

        Args:
            nodeid (str): the name of the node we are publishing the target to
            channels (list of str): list of channel names for the node we are announcing targets to
            vmag (list of float): the 'V' target to be set for each channel
            vang (list of float): the 'delta' target to be set for each channel
            kvbases (list of float or None): the KV base for each channel
            kvabases (list of float or None): the KVA base for each channel
        """
        self._log.info(f"SPBC announcing channels {channels}, vmag {vmags}, vang {vangs} to node {nodeid}")
        # wrap value in nullable Double if provided

        targets = []
        for idx, channel in enumerate(channels):
            kvbase = Double(value=kvbases[idx]) if kvbases is not None else None
            kvabase = Double(value=kvabases[idx]) if kvabases is not None else None
            targets.append(
                EnergisePhasorTarget(
                    nodeID=nodeid,
                    channelName=channels[idx],
                    angle=vangs[idx],
                    magnitude=vmags[idx], 
                    kvbase=kvbase,
                    KVAbase=kvabase,
                )
            )

        await self.publish(self.namespace, f"spbc/{self.name}/node/{nodeid}", XBOS(
            EnergiseMessage = EnergiseMessage(
            SPBC=SPBC(
                time=int(datetime.utcnow().timestamp()*1e9),
                phasor_targets=targets
                )
            )))


class LPBCProcess(XBOSProcess):
    """
    Wrapper process for local phasor-based control in Python.
    Requires a uPMU
    """
    def __init__(self, cfg):
        super().__init__(cfg)
        if 'namespace' not in cfg:
            raise ConfigMissingError('namespace')
        self.namespace = b64decode(cfg['namespace'])
        if 'local_channels' not in cfg:
            raise ConfigMissingError('local_channels')
        if 'reference_channels' not in cfg:
            raise ConfigMissingError('reference_channels')
        if 'name' not in cfg:
            raise ConfigMissingError('name')
        if 'spbc' not in cfg:
            raise ConfigMissingError('spbc')
        if 'rate' not in cfg:
            raise ConfigMissingError('rate')

        # locks
        self._reference_phasor_lock = asyncio.Lock()
        self._local_phasor_lock = asyncio.Lock()

        self.local_channels = cfg['local_channels']
        self.reference_channels = cfg['reference_channels']

        self.name = cfg['name']
        self.spbc = cfg['spbc']
        self._rate = int(cfg['rate'])

        # local buffers for phasor data
        # key: channel name, value: phasor data
        self.local_phasor_data = {}
        self.reference_phasor_data = {}
        self.last_spbc_command = None
        self.control_on = False

        for local_channel in self.local_channels:
            self.local_phasor_data[local_channel] = []
            cb = partial(self._local_upmucb, local_channel)
            schedule(self.subscribe_extract(self.namespace, f"upmu/{local_channel}", ".C37DataFrame", cb, f"local_channel_{local_channel}"))
        for reference_channel in self.reference_channels:
            self.reference_phasor_data[reference_channel] = []
            cb = partial(self._reference_upmucb, reference_channel)
            schedule(self.subscribe_extract(self.namespace, f"upmu/{reference_channel}", ".C37DataFrame", cb, f"reference_phasor_{reference_channel}"))

        # TODO: listen to SPBC
        print(f"spbc/{self.spbc}/node/{self.name}")
        schedule(self.subscribe_extract(self.namespace, f"spbc/{self.spbc}/node/{self.name}", ".EnergiseMessage.SPBC", self._spbccb, "spbc_sub"))
	
        schedule(self.call_periodic(self._rate, self._trigger, runfirst=False))

        self._log.info(f"initialized LPBC: {cfg}")

    def _local_upmucb(self, channel, resp):
        """Stores the most recent local upmu reading"""
        frame = resp.values[-1]['phasorChannels'][0]
        #self._log.info(f"got {len(frame['data'])} values on local")
        if channel not in self.local_phasor_data:
            self.local_phasor_data[channel] = []
        self.local_phasor_data[channel].extend(frame['data'])

    def _reference_upmucb(self, channel, resp):
        """Stores the most recent reference upmu reading"""
        if len(resp.values) == 0:
            self._log.error("no content in UPMU message")
            return
        if len(resp.values[-1]['phasorChannels']) == 0:
            self._log.error("no phasor channels in UPMU message")
            return
        frame = resp.values[-1]['phasorChannels'][0]
        if channel not in self.reference_phasor_data:
            self.reference_phasor_data[channel] = []
        self.reference_phasor_data[channel].extend(frame['data'])

    def _spbccb(self, resp):
        """Stores the most recent SPBC command"""
        if len(resp.values) == 0:
            self._log.error("no content in SPBC message")
            return
        resp = resp.values[-1]
        resp['phasor_targets'] = resp.pop('phasorTargets')
        self.last_spbc_command = resp

    def _received_local_phasor_data(self):
        return sum(map(len, self.local_phasor_data.values())) > 0
    def _received_reference_phasor_data(self):
        return sum(map(len, self.reference_phasor_data.values())) > 0

    async def _trigger(self):
        try:
            if not self._received_local_phasor_data():
                self._log.warning(f"LPBC {self.name} has not received a local upmu reading")
                #return
            if not self._received_reference_phasor_data():
                self._log.warning(f"LPBC {self.name} has not received a reference upmu reading")
                #return
            if self.last_spbc_command is None:
                self._log.warning(f"LPBC {self.name} has not received an SPBC command")
                #return
            #spbc_cmd = self.last_spbc_command.values[-1] # most recent
            #targets = spbc_cmd['phasor_targets']
            async with self._local_phasor_lock:
                local_phasors = [self.local_phasor_data.pop(local_channel) for local_channel in self.local_channels]
            async with self._reference_phasor_lock:
                reference_phasors = [self.reference_phasor_data.pop(reference_channel) for reference_channel in self.reference_channels]

            phasor_targets = self.last_spbc_command

            # rebuild buffers
            async with self._local_phasor_lock:
                for local_channel in self.local_channels:
                    self.local_phasor_data[local_channel] = []

            async with self._reference_phasor_lock:
                for reference_channel in self.reference_channels:
                    self.reference_phasor_data[reference_channel] = []

            await self.do_trigger(local_phasors, reference_phasors, phasor_targets)
        except Exception as e:
            self._log.error(f"Error in processing trigger: {traceback.format_exc()}")

    async def do_trigger(self, local_phasors, reference_phasors, phasor_targets):
        self._log.info(f"""LPBC {self.name} received call at {datetime.now()}:
    Local phasor has {len(local_phasors)} channels
    Reference phasor has {len(reference_phasors)} channels
    SPBC targets: {phasor_targets}
        """)
        try:
            status = self.step(local_phasors, reference_phasors, phasor_targets)
        except Exception as e:
            print(">>>",e)
        if status is None:
            return

        for required in ['p_max','q_max','phases','phasor_errors','p_saturated','q_saturated']:
            if required not in status:
                raise Exception(f"Need {required} key in status dictionary")

        statuses = []
        for idx, phase_name in enumerate(status.pop('phases')):
            p_max = Double(value=status['p_max'][idx]) if status['p_max'][idx] else None
            q_max = Double(value=status['q_max'][idx]) if status['q_max'][idx] else None
            channel_status = ChannelStatus(
                nodeID=self.name,
                channelName=phase_name,
                phasor_errors=Phasor(
                    magnitude=status['phasor_errors']['V'][idx],
                    angle=status['phasor_errors']['delta'][idx],
                ),
                p_saturated=status['p_saturated'][idx],
                q_saturated=status['q_saturated'][idx],
                p_max=p_max,
                q_max=q_max,
            )
            statuses.append(channel_status)

        msg = XBOS(
                EnergiseMessage=EnergiseMessage(
                    LPBCStatus=LPBCStatus(
                        time=int(datetime.utcnow().timestamp()*1e9),
                        statuses=statuses,
                    )
                ))
        await self.publish(self.namespace, f"lpbc/{self.name}", msg)



from pyxbos.driver import *
from pyxbos import hvac_pb2
import pymortar
import BAC0
import re
import os,sys
import json
import time

class BACnetDriver(Driver):
    def setup(self, cfg):
        self.pymortar_client = pymortar.Client({
            'username': cfg['mortar_api_username'],
            'password': cfg['mortar_api_password']
        })
        self.bacnet = BAC0.connect()
        # A dictionary mapping BACnet point names to their Brick classes grouped by Equipment class
        self.point_class_dict = get_point_class_dict(self.pymortar_client, cfg['building'])
        # A list of all the BACnet devices in the network
        self.devices = [BAC0.device(dev[2], dev[3], self.bacnet) for dev in self.bacnet.devices]

    def read(self, requestid=None):
        for dev in self.devices:
            # A dictionary mapping BACnet point names mapped their respective values
            point_value_map = get_point_value_dict(dev.points)

            #AHU-1.DPR-O Damper point MISSING from BACnet points for orinda-public-library

            # Output will look like { "AHU": { "AHU2-1": { "return_air_temperature_sensor": 67.47, ..... }, ....}, .....}
            output = {}
            for equipclass in self.point_class_dict:
                if equipclass not in output:
                    output[equipclass] = {}

                for point in self.point_class_dict[equipclass]:
                    pieces = point.split(".")
                    new_point = point
                    if len(pieces) > 2:
                        # This is to fix point names like "AHU-2.SF-.SF-O" to "AHU-2.SF-O"
                        new_point = pieces[0] + "." + pieces[-1]

                    if "DPRPOS" in point:
                        # Pymortar may return point names with DPRPOS while BACnet's naming convention is DMPRPOS
                        # This fixes those points to follow BACnet's naming convention
                        new_point = pieces[0] + "." + "DMPRPOS"

                    if new_point in point_value_map:
                        # Gets the Brick class of the point name using the point to class name dict
                        point_class = self.point_class_dict[equipclass][point]['point_class']
                        
                        if pieces[0] not in output[equipclass]:
                            output[equipclass][pieces[0]] = {}
                        
                        output[equipclass][pieces[0]][point_class.lower()] = point_value_map[new_point]
            
            # Goes through every equipment class in output, and sends out a xbos_pb2 message
            for equipclass in output:
                # Maps all properties of each equipment to their respective BACnet values
                # Looks like { "time": 1564278587100909056, ahu": hvac_pb2.AHU(return_air_temperature_sensor = 67.47, ....), .... }
                device_state_data = {}
                # all_eq looks like { "AHU2-1": {...}, "AHU2-3": {....}, .... }
                all_eq = output[equipclass]
                for eq in all_eq:
                    # Dictionary of properties for this piece of equipment
                    # Looks like { "return_air_temperature_sensor": 67.47, .... }
                    props = {}
                    for prop in all_eq[eq]:
                        # For each property like "return_air_temperature_sensor"
                        if prop not in ['vvt', 'vav']:
                            # Gets the value of that property or sets it to None if it doesn't exist
                            value = all_eq[eq].get(prop, None)

                            if not isinstance(value, str):
                                # Changes the value into a Double type (defined in the proto file) if not a string
                                # If the point_value_dict was created correctly, then all values should be converted into doubles already
                                props[prop] = types.Double(value=value)

                    # e.g. Calls hvac_pb2.AHU(return_air_temperature_sensor = 67.47)
                    device_state_data[equipclass.lower()] = getattr(hvac_pb2, equipclass)(**props)
                    device_state_data['time'] = int(time.time()*1e9)

                    msg = xbos_pb2.XBOS(XBOSIoTDeviceState = iot_pb2.XBOSIoTDeviceState(**device_state_data))
                    self.report(eq, msg)

def get_point_value_dict(device_points):
    """ Creates a dictionary of BACnet point names mapped to their respective values

    e.g. { "AHU-1.SF-A": false, "VVT-4.MAXDMP": 75, "Garage.ZN2-Q": 1.77, .... }

    :param device_points: List of all the point names in a particular BACnet device
    :return: Dictionary of BACnet point names mapped to their respective values
    """
    points = [str(point).strip() for point in device_points]
    point_value_dict = {}

    for point in points:
        point_name, point_value = point.split(" : ")

        point_name = point_name.split(".", 1)[1]

        if point_value.lower() in ["true", "occupied", "cooling"]:
            # If the value is "occupied" or "cooling" then, converting the value to a boolean True to
            # make it easier to track
            point_value_dict[point_name] = True
        elif point_value.lower() in ["false", "unknown", "unoccupied"]:
            # If the value is "unknown" or "unoccupied" then, converting the value to a boolean False to
            # make it easier to track
            point_value_dict[point_name] = False
        elif any(char.isdigit() for char in point_value):
            # If the BACnet value contains any numbers, get rid of all 
            # non-digit characters and convert to float
            point_value_dict[point_name] = float(re.sub(r'[^0-9.]','', point_value))
        else:
            point_value_dict[point_name] = point_value

    return point_value_dict

def get_point_class_dict(pymortar_client, building):
    """ Creates a dictionary of BACnet point names mapped to their respective Brick classes.
    Points are also grouped by Equipment class. Point names and classes are obtained from Pymortar. 

    e.g. { "AHU": { 
            "AHU-3.RA-T": {
                "point_name": "AHU-3.RA-T",
                "point_class": "Return_Air_Temperature_Sensor"
            }, .....,
          "VAV": { 
            .....
           }
        }

    :param pymortar_client: Pymortar client
    :param building: Name of building that is being analyzed
    :return: Dictionary of BACnet point names mapped to their respective Brick classes grouped by Equipment class
    """
    v = pymortar.View(
            name="equipment",
            definition="""
            SELECT ?equipname ?equipclass ?point ?pointclass FROM %s WHERE {
                ?equipclass rdfs:subClassOf+ brick:Equipment .
                ?equipname rdf:type ?equipclass .
                ?equipname bf:hasPoint ?point .
                ?point rdf:type ?pointclass
            };""" % building
        )

    res = pymortar_client.fetch(pymortar.FetchRequest(
            sites=[building],
            views=[v]
        ))

    eq_view = res.view('equipment')

    point_class_dict = {}
    groups = eq_view.groupby('equipclass').groups
    # Equipment classes that are not being considered and will be excluded from the dictionary
    # Add to this list of excluded equipment for other buildings
    excluded_equipment = ['Boiler', 'Chilled_Water_Pump', 'Hot_Water_Pump']
    # Clusters all points with subclasses of Fan and Damper (i.e. Exhaust_Fan)
    clustered_equipment = dict.fromkeys(["Damper", "Fan"], None)
    clustered = None

    for equipclass, indexes in groups.items():
        if equipclass not in excluded_equipment:
            for clustered_eq in clustered_equipment:
                # Union all the clustered equipment
                if clustered_eq.lower() in equipclass.lower():
                    clustered = clustered_eq
                    break

            if clustered:
                if clustered not in point_class_dict:
                    point_class_dict[clustered] = {}
            else:
                if equipclass not in point_class_dict:
                    point_class_dict[equipclass] = {}

            points = eq_view.iloc[indexes][['point','pointclass']].values
            for p in points:
                point_name, point_class = p[0], p[1]
                if clustered:
                    point_class_dict[clustered][point_name] = { "point_name": point_name, "point_class": point_class}
                else:
                    point_class_dict[equipclass][point_name] = { "point_name": point_name, "point_class": point_class}

            clustered = None

    return point_class_dict

if __name__ == '__main__':
    logging.basicConfig(level="INFO", format='%(asctime)s - %(name)s - %(message)s')

    building = 'orinda-public-library'

    cfg = {
        'wavemq': 'localhost:4516',
        'waved': 'localhost:410',
        'namespace': 'GyCetklhSNcgsCKVKXxSuCUZP4M80z9NRxU1pwfb2XwGhg==',
        'base_resource': 'sriharsha',
        'entity': 'sriharsha.ent',
        'id': 'pyxbos-driver-hvac-1',
        'rate': 10,
        'building': building,
        'mortar_api_username': os.environ['MORTAR_API_USERNAME'],
        'mortar_api_password': os.environ['MORTAR_API_PASSWORD']
    }

    current_driver = BACnetDriver(cfg)
    current_driver.begin()